@@ -22,17 +22,24 @@
 #define PAMAUDIT_LOGGED 1
 
 static int
-_pam_audit_writelog(pam_handle_t *pamh, int audit_fd, int type, 
+_pam_audit_writelog(pam_handle_t *pamh, int audit_fd, int type,
 	const char *message, int retval)
 {
+  static int old_errno = -1;
   int rc;
-  char buf[256];
+  char buf[32];
 
-  snprintf(buf, sizeof(buf), "PAM: %s acct=%s ", message, 
-	(retval != PAM_USER_UNKNOWN && pamh->user) ? pamh->user : "?");
+  snprintf(buf, sizeof(buf), "PAM:%s", message);
 
-  rc = audit_log_user_message( audit_fd, type, buf,
-        pamh->rhost, NULL, pamh->tty, retval == PAM_SUCCESS );
+  rc = audit_log_acct_message (audit_fd, type, NULL, buf,
+       (retval != PAM_USER_UNKNOWN && pamh->user) ? pamh->user : "?",
+	-1, pamh->rhost, NULL, pamh->tty, retval == PAM_SUCCESS );
+
+  if (rc == -1 && errno != old_errno)
+    {
+      old_errno = errno;
+      pam_syslog (pamh, LOG_CRIT, "audit_log_acct_message() failed: %m");
+    }
 
   pamh->audit_state |= PAMAUDIT_LOGGED;
   return rc;
@@ -59,18 +66,18 @@ _pam_auditlog(pam_handle_t *pamh, int action, int retval, int flags)
     pam_syslog(pamh, LOG_CRIT, "audit_open() failed: %m");
     return PAM_SYSTEM_ERR;
   }
-                          
+
   switch (action) {
   case PAM_AUTHENTICATE:
     message = "authentication";
     type = AUDIT_USER_AUTH;
     break;
   case PAM_OPEN_SESSION:
-    message = "session open";
+    message = "session_open";
     type = AUDIT_USER_START;
     break;
   case PAM_CLOSE_SESSION:
-    message = "session close";
+    message = "session_close";
     type = AUDIT_USER_END;
     break;
   case PAM_ACCOUNT:
@@ -105,7 +112,7 @@ _pam_auditlog(pam_handle_t *pamh, int action, int retval, int flags)
 
   if (_pam_audit_writelog(pamh, audit_fd, type, message, retval) < 0)
     retval = PAM_SYSTEM_ERR;
-  
+
   audit_close(audit_fd);
   return retval;
 }
